
(* print_string is just a pre-defined function (not a keyword or anything) *)
let x = print_string "Hello, World!\n"

(*same as previous with nothing bound to () -- underscore is not a variable *)
let _ = print_string "Hello, World!\n"

(*same w/ variables and infix concat function*)
let h = "Hello, "
let w = "World!\n"
let _ = print_string (h ^ w)

(*function f: ignores its argument & prints*)
let f x = print_string (h ^ w)

(*so these both print (call is juxtapose)*)
let y1 = f 37
let y2 = f f (* pass function itself *)
(*but this does not (y1 bound to ())*)
let y3 = y1

(* print_string : string->unit,  : string *)
let x = print_string "Hello, World!\n"
(* x: unit *)

(* ^ : string->string->unit *)
let f x = print_string (h ^ w) 
(* f : a -> unit *)
let y1 = f 37 (* y1 : unit *) 
let y2 = f f  (* y2 : unit *) 
let y3 = y1   (* y3 : unit *)

(* constraining a type *)
let g (x:int) = f x

let _ = g 37 
(*let _ = g "hi" (*no typecheck, but f "hi" does*) *)

(* A really small infinite loop -- do not worry about the inferred type *)
(* rec is necesary -- else forever is unbound (would refer to earlier one) *)
let rec forever x = forever x

(* A more useful recursive function, incorrect for negative x *)
(* Note precedence requires parentheses here *)
let rec fact x = if x==0 then 1 else x * (fact(x-1))

(*everything an expression, for example  if e1 then e2 else e3*) 
let fact2 x = 
  (if x==0 then 1 else x * (fact(x-1))) * 2 / 2

(* local variables and functions look very similar to top-level ones *)
(* indentation is just style *)
(* If e1 and e2 are expressions, then let x = e1 in e2 is an expression *)
let quadruple x = 
  let double y = y + y in
  let ans = double x + double x in
  ans

let _ = print_string((string_of_int(quadruple 7)) ^ "\n")

(* functions can be anonymous *)
let quadruple2 x = (fun x -> x + x) x + (fun x -> x + x) x

(* In fact, we've been using a shortcut:
   "let f x = e" is really just "let f = fun x -> e" *)
let quadruple3 x = 
  let double = fun x -> x + x in
  double x + double x

let _ = print_string ((string_of_int (quadruple  7)) ^ "\n")
let _ = print_string ((string_of_int (quadruple2 7)) ^ "\n")
let _ = print_string ((string_of_int (quadruple3 7)) ^ "\n")

(* with "boring" sharing (fine here) *)
let print_i_nl i = print_string ((string_of_int i) ^ "\n")

(* note: e1; e2 is just let _ = e1 in e2, 
         i.e., evaluate e1, then evaluate e2, result of e2 is result *)
let _ = print_i_nl (quadruple  7);
        print_i_nl (quadruple2 7);
        print_i_nl (quadruple3 7)

(* With passing functions... (and our first 2-argument function) *)
(* and useful polymorphism -- though we're not using it *)
let print_i_nl2 i f = print_i_nl (f i)
let _ = print_i_nl2 7 quadruple ;
        print_i_nl2 7 quadruple2;
        print_i_nl2 7 quadruple3

let print_on_seven1 f = print_i_nl2 7 f

let print_on_seven2 = print_i_nl2 7

(* What's really going on:
   * Every function takes exactly one argument
   * "Multiple arguments is syntactic sugar"
   * Application "associates to the left"
 *)

(* 2 ways to write the same thing *)
let print_i_nl2 i f = print_i_nl (f i)
let print_i_nl2 = fun i -> (fun f -> print_i_nl (f i))
(*print_i_nl2 : (int -> ((int -> int) -> unit))
  i.e.,         (int -> (int -> int)  -> unit)
*)

(* 2 ways to write the same thing *)
let _ = print_i_nl2 7 quadruple
let _ = (print_i_nl2 7) quadruple

(* closures and static scoping *)
let y = 5
let return11 = (* unit -> int *)
	 let x = 6 in
   fun () -> x + y
let y = 7
let x = 8
let _ = print_string "demonstrate closures..."
let _ = print_i_nl (return11 ()) (*prints 11 !!!*)


let x = 9
let f () = x+1
let x = x+1
let g () = x+1
let _ = print_i_nl (f() + g())

(* Variables are not mutable *)
let x = 9
let f() = x
let x = x + 1
let g() = x
let _ = print_i_nl (f()+g())

(* Refs are mutable, but should be used very very sparingly. *)
let x = ref 9
let f () = !x
let _ = x := !x + 1
let g () = !x
let _ = print_i_nl (f()+g())

(************************ Aggregate Types ******************************)

(* records and polymorphic records *)
type my_int_pair = {first : int; second : int}
let sum_int_pr x = x.first + x.second
let pr1 = {first = 3; second = 4}
let _ = sum_int_pr pr1 + sum_int_pr {first=5;second=6}

type 'a mypair = {a_first : 'a; a_second : 'a}
let sum_pr f x = f x.a_first + f x.a_second
let pr2 = {a_first = 3; a_second = 4}(*int pair*)
let _ = sum_int_pr pr1 + sum_pr (fun x->x) {a_first=5;a_second=6}

let pr3 = {a_first = "hi"; a_second = "mom"}
let pr4 = {a_first = pr2;  a_second = pr2}
let sum_int      = sum_pr (fun x -> x)
let sum_str      = sum_pr String.length
let sum_int_pair = sum_pr sum_int
let _ = print_i_nl (sum_int pr2)
let _ = print_i_nl (sum_str pr3)
let _ = print_i_nl (sum_int_pair pr4)

(* datatypes *)
type food = Foo of int | Bar of my_int_pair | Baz of int * int | Quux

let foo3     = Foo (1 + 2)
let bar12    = Bar pr1
let baz1_120 = Baz(1,fact 5)
let quux     = Quux (* not much point in this *)

let is_a_foo x =
  match x with (* better than "downcasts" *)
   Foo i    -> true
 | Bar pr   -> false
 | Baz(i,j) -> false
 | Quux     -> false
(* Note: variant names must be capitalized, other identifiers must not be. *)

(* booleans and if are built-in funny syntax (uncapitalized) *)

let print_bool1 b = 
  print_string (if b then "true\n" else "false\n")
let print_bool2 b = 
  print_string (match b with true -> "true\n" | false -> "false\n")

let _ = print_bool1 true; print_bool2 false

(* recursive and polymorphic datatypes *)
type int_tree = Leaf | Node of int * int_tree * int_tree
type 'a lst = Null | Cons of 'a * 'a lst

let lst1 = Cons(3,Null)
let lst2 = Cons(1,Cons(2,lst1))
(* let lst_bad = Cons("hi",lst2) *)
let lst3 = Cons("hi",Cons("mom",Null))
let lst4 = Cons (Cons (3,Null), Cons (Cons (4,Null), Null))

let rec len lst = (* a lst -> int *)
  match lst with
   Null -> 0
 | Cons(x,rest) -> 1 + len rest

let rec sum lst = (* int lst -> int *)
  match lst with
   Null -> 0
 | Cons(x,rest) -> x + sum rest

let rec append lst1 lst2 = (* a lst -> a lst -> int *)
  match lst1 with
    Null -> lst2
  | Cons(x,rest) -> Cons(x, append rest lst2)

let rec append2 lst1 lst2 = (* built-in infix @ *)
  match lst1 with
    [] -> lst2
  | x::rest -> x :: append2 rest lst2

(* tuples *)
let x = (3,"hi",(fun x -> x), fun x -> x ^ "ism")
let z = match x with (i,s,f1,f2) -> f1 i
let z = (let (i,s,f1,f2) = x in f1 i)

(****** higher-order functions (more later) ****)

let rec mymap f lst =
  match lst with
    [] -> []
  | hd::tl -> (f hd)::(mymap f tl)

let lst234 = mymap (fun x -> x+1) [1;2;3]

let lst345 = List.map (fun x -> x+1) [2;3;4]

let incr_list = mymap (fun x -> x+1)

(* Using standard libraries: Modulename.functionName
   In homework 1: List, String, Char (we point you to functions)
*)

(****** fancier pattern-matching ******)

type sign = P | N | Z

let multsign x1 x2 = 
  let sign x = 
    if x>=0 then (if x=0 then Z else P) else N
  in
  match (sign x1,sign x2) with
    (P,P) -> P
  | (N,N) -> N
  | (Z,_) -> Z
  | (_,Z) -> Z
  | _     -> N (* many say bad style! *)

exception ZipLengthMismatch

let rec zip3 lst1 lst2 lst3 = 
 match (lst1,lst2,lst3) with
    ([],[],[]) -> []
  | (hd1::tl1,hd2::tl2,hd3::tl3) -> 
		(hd1,hd2,hd3)::(zip3 tl1 tl2 tl3)
  | _ -> raise ZipLengthMismatch

(* in other files: 
   1. creating modules
   2. closure idioms 
*)
